/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Mar  3 2016
 */
#include <errno.h>
#include <inttypes.h>
#include <string.h>

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

#define json_skip_whitespace(json) (json) += strspn(json, " \t\n\r")
#define JSON_MINBUF	128

static const char tok_null[] = "null";
static const char tok_true[] = "true";
static const char tok_false[] = "false";

static int	jgetstring(char *dst, const char **json, char *emsg,
                           size_t emsgsize);
static int	i2j_value(const char *key, const mxArray *value,
                          const mxArray *format, char **json,
                          char **jend, size_t *jlen, char *emsg,
                          size_t emsgsize);
static int	json_bufcat(char **buf, char **end, size_t *len, int quoted,
                            const char *data);


/* --- j2s_tokerror -------------------------------------------------------- */

/** Append an error message with the current token
 */
static void *
j2s_tokerror(const char *what, const char *json, char *emsg, size_t emsgsize)
{
  char tok[128];
  size_t s;

  s = strcspn(json, " \t\n\r{}[],:\"");
  if (!s) s = 1;
  if (s > sizeof(tok)-1) s = sizeof(tok)-1;
  strncpy(tok, json, s);

  snprintf(emsg, emsgsize, "unexpected %s \"%s\"", what, tok);
  return NULL;
}


/* --- j2s_value ----------------------------------------------------------- */

/** Parse a JSON value
 */
mxArray *
j2s_value(const char **json, char *emsg, size_t emsgsize)
{
  json_skip_whitespace(*json);
  switch(**json) {
    case 0:
      return NULL;

    case '{': /* object */ {
      mxArray **va = NULL;
      mxArray *v, *s;
      char key[256];
      size_t na = 0;
      int i, m, isarray;

      ++(*json);
      json_skip_whitespace(*json);

      s = mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
      if (!s) return s;
      m = isarray = 0;

      while(**json != '}') {
        /* key: string */
        if (**json != '"')
          return j2s_tokerror("non-string key", *json, emsg, emsgsize);
        ++(*json);
        if (jgetstring(key, json, emsg, emsgsize)) return NULL;

        if (isarray || (key[0] >= '0' && key[0] <= '9')) {
          isarray = 1;
          i = atoi(key);
          if (i > m) m = i;
        } else
          i = mxAddField(s, key);

        json_skip_whitespace(*json);
        if (**json != ':')
          return j2s_tokerror("element separator", *json, emsg, emsgsize);
        ++(*json);

        /* value */
        if (isarray) {
          if (i >= na) {
            size_t ona = na;
            if (na) na <<= 1; else na = 16;
            va = realloc(va, na * sizeof(*va));
            if (!va) return NULL;
            memset(&va[ona], 0, (char *)&va[na] - (char *)&va[ona]);
          }
          if (!(va[i] = j2s_value(json, emsg, emsgsize))) return NULL;
        } else {
          if (!(v = j2s_value(json, emsg, emsgsize))) return NULL;
          mxSetFieldByNumber(s, 0, i, v);
        }

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != '}')
          return j2s_tokerror("element", *json, emsg, emsgsize);
      }
      ++(*json);

      if (isarray) {
        mxDestroyArray(s);
        m++;
        s = mxCreateCellArray(2, (mwSize[2]){1, m});
        if (s) while(m--) if (va[m]) mxSetCell(s, m, va[m]);
        free(va);
      }

      return s;
    }

    case '[': /* array */ {
      mxArray *a, **v;
      size_t i, n;

      ++(*json);
      json_skip_whitespace(*json);

      i = 0;
      n = 16;
      v = malloc(n * sizeof(*v));
      if (!v) return NULL;
      while(**json != ']') {
        if (i >= n) {
          n <<= 1;
          v = realloc(v, n * sizeof(*v));
          if (!v) return NULL;
        }
        if (!(v[i] = j2s_value(json, emsg, emsgsize))) return NULL;

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != ']')
          return j2s_tokerror("element", *json, emsg, emsgsize);

        i++;
      }
      ++(*json);

      a = mxCreateCellArray(2, (mwSize[2]){1, i});
      if (a) while(i--) mxSetCell(a, i, v[i]);
      free(v);
      return a;
    }

    case '"': /* string */ {
      char key[256];

      ++(*json);
      if (jgetstring(key, json, emsg, emsgsize)) return NULL;
      return mxCreateString(key);
    }

    case 't': /* bare word: true, false, null */
    case 'f':
    case 'n':
      if (!strncmp(*json, tok_true, sizeof(tok_true)-1)) {
        *json += sizeof(tok_true)-1;
        return mxCreateLogicalScalar(true);
      }
      if (!strncmp(*json, tok_false, sizeof(tok_false)-1)) {
        *json += sizeof(tok_false)-1;
        return mxCreateLogicalScalar(false);
      }
      if (!strncmp(*json, tok_null, sizeof(tok_null)-1)) {
        *json += sizeof(tok_null)-1;
        return mxCreateCellArray(1, &(mwSize){0});
      }

      return j2s_tokerror("value", *json, emsg, emsgsize);

    default: /* number or error */ {
      const char *start = *json;
      char *end;
      double v;

      v = strtod(start, &end);
      if (end == start) return j2s_tokerror("value", start, emsg, emsgsize);
      *json = end;

      return mxCreateDoubleScalar(v);
    }
  }

  snprintf(emsg, emsgsize, "internal error");
  return NULL;
}

static int
jgetstring(char *dst, const char **json, char *emsg, size_t emsgsize)
{
  const char *s;
  char *d;

  s = *json;
  d = dst;
  while (*s != '"' && *s && d - dst < 255) {
    *d = *s++;
    if (*d == '\\') {
      *d = *s++;
      if (!*d) break;
    }
    d++;
  }
  if (*s != '"') {
    j2s_tokerror("end of string", *json, emsg, emsgsize);
    return -1;
  }

  *d = 0;
  *json = s+1;
  return 0;
}


/* --- i2j_value ----------------------------------------------------------- */

/** Build a JSON value
 */
char *
s2j_input(const mxArray *input, char *emsg, size_t emsgsize)
{
  mxArray *data, *format, *k, *v;
  const char *field;
  char *json, *end;
  size_t len;
  int i, n, s;

  data = mxGetField(input, 0, "data");
  format = mxGetField(input, 0, "format");
  if (!data || !format) {
    snprintf(emsg, emsgsize, "internal error");
    return NULL;
  }

  len = JSON_MINBUF;
  json = end = malloc(len);
  if (!json) {
    snprintf(emsg, emsgsize, "no memory");
    return NULL;
  }

  if ((s = json_bufcat(&json, &end, &len, 0, "{"))) goto error;
  n = mxGetNumberOfFields(format);
  for(i = 0; i < n; i++) {
    field = mxGetFieldNameByNumber(format, i);
    v = mxGetFieldByNumber(format, 0, i);
    k = mxGetField(data, 0, field);
    if (!k) {
      snprintf(emsg, emsgsize, "missing input value for argument `%s'", field);
      goto error;
    }

    if (i && json_bufcat(&json, &end, &len, 0, ",")) goto error;
    if (json_bufcat(&json, &end, &len, 0, "\"")) goto error;
    if (json_bufcat(&json, &end, &len, 0, field)) goto error;
    if (json_bufcat(&json, &end, &len, 0, "\":")) goto error;

    s = i2j_value(field, k, v, &json, &end, &len, emsg, emsgsize);
    if (s) goto error;
  }
  if (json_bufcat(&json, &end, &len, 0, "}")) goto error;
  return json;

error:
  free(json);
  return NULL;
}


static int
i2j_value(const char *key, const mxArray *value, const mxArray *format,
          char **json, char **jend, size_t *jlen, char *emsg, size_t emsgsize)
{
  char kind[64], skey[256];
  const char *field;
  mxArray *type, *stype, *svalue;
  mxArray *v;
  double *l;
  char *t;
  int i, n, s;

  v = mxGetField(format, 0, "kind");
  if (!v) goto fatal;
  mxGetString(v, kind, sizeof(kind));

  if (!strcmp(kind, "struct") || !strcmp(kind, "union")) {
    if (!mxIsStruct(value)) {
      snprintf(emsg, emsgsize, "`%s' input value must be a structure", key);
      return -1;
    }
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    /* check missing arguments */
    n = mxGetNumberOfFields(type);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(type, i);
      v = mxGetField(value, 0, field);
      if (!v) {
        snprintf(emsg, emsgsize, "missing input argument `%s.%s'", key, field);
        return -1;
      }
    }

    /* build json */
    if ((s = json_bufcat(json, jend, jlen, 0, "{"))) return s;
    n = mxGetNumberOfFields(value);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(value, i);
      svalue = mxGetFieldByNumber(value, 0, i);
      snprintf(skey, sizeof(skey), "%s.%s", key, field);
      stype = mxGetField(type, 0, field);
      if (!stype) {
        snprintf(emsg, emsgsize, "unknown input argument `%s'", skey);
        return -1;
      }

      if (i && (s = json_bufcat(json, jend, jlen, 0, ","))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, field))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, "\":"))) return s;

      s = i2j_value(skey, svalue, stype, json, jend, jlen, emsg, emsgsize);
      if (s) return s;
    }
    if ((s = json_bufcat(json, jend, jlen, 0, "}"))) return s;
    return 0;

  } else if (!strcmp(kind, "array") || !strcmp(kind, "sequence")) {
    if (!mxIsCell(value)) {
      snprintf(emsg, emsgsize, "%s %s expects cell array data", kind, key);
      return -1;
    }
    n = mxGetN(value);

    /* check missing arguments */
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    v = mxGetField(format, 0, "length");
    if (v) {
      l = mxGetPr(v);
      if (!l || n != (size_t)*l) {
        snprintf(emsg, emsgsize, "%s %s has wrong size: must be %zu elements",
                 kind, key, l?(size_t)*l:n);
        return -1;
      }
    }

    /* build json */
    if ((s = json_bufcat(json, jend, jlen, 0, "["))) return s;
    for(i = 0; i < n; i++) {
      svalue = mxGetCell(value, i);
      snprintf(skey, sizeof(skey), "%s[%d]", key, i);

      if (i && (s = json_bufcat(json, jend, jlen, 0, ","))) return s;

      s = i2j_value(skey, svalue, type, json, jend, jlen, emsg, emsgsize);
      if (s) return s;
    }
    if ((s = json_bufcat(json, jend, jlen, 0, "]"))) return s;

    return 0;

  } else if (!strcmp(kind, "optional")) {
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    if (mxIsEmpty(value)) {
      s = json_bufcat(json, jend, jlen, 0, tok_null);
    } else {
      s = i2j_value(key, value, type, json, jend, jlen, emsg, emsgsize);
    }
    return s;

  } else if (!strcmp(kind, "char") || !strcmp(kind, "string")) {
    if (!mxIsChar(value)) {
      snprintf(emsg, emsgsize, "%s %s expects string data", kind, key);
      return -1;
    }
    t = mxArrayToString(value);
    if (!t) goto fatal;
    n = strlen(t);

    v = mxGetField(format, 0, "length");
    if (v) {
      l = mxGetPr(v);
      if (l && n > (size_t)*l) {
        snprintf(emsg, emsgsize, "%s %s too long: length %d, %zu max",
                 kind, key, n, (size_t)*l);
        return -1;
      }
    }

    if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
    if ((s = json_bufcat(json, jend, jlen, 1, t))) return s;
    if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;

    mxFree(t);
    return 0;

  } else if (!strcmp(kind, "enum")) {
    if (!mxIsChar(value)) {
      snprintf(emsg, emsgsize, "%s %s expects string data", kind, key);
      return -1;
    }
    t = mxArrayToString(value);
    if (!t) goto fatal;

    v = mxGetField(format, 0, "values");
    if (!v) goto fatal;
    n = mxGetN(v);

    for(i = 0; i < n; i++) {
      svalue = mxGetCell(v, i);
      mxGetString(svalue, skey, sizeof(skey));
      if (!strncmp(skey, t, sizeof(skey))) {
        if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
        if ((s = json_bufcat(json, jend, jlen, 0, t))) return s;
        if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
        mxFree(t);
        return 0;
      }
    }

    snprintf(emsg, emsgsize, "bad value for %s %s: %s", kind, key, t);
    mxFree(t);
    return -1;

  } else {
    void *d;

    if (!mxIsNumeric(value)) {
      snprintf(emsg, emsgsize, "bad numeric value for %s %s", kind, key);
      return -1;
    }
    n = mxGetNumberOfElements(value);
    if (n != 1) {
      snprintf(emsg, emsgsize, "bad dimension %d for %s %s", n, kind, key);
      return -1;
    }
    d = mxGetData(value);
    if (!d) goto fatal;

    switch(mxGetClassID(value)) {
      case mxINT8_CLASS:
        snprintf(skey, sizeof(skey), "%" PRId8, *(int8_t *)d); break;
      case mxINT16_CLASS:
        snprintf(skey, sizeof(skey), "%" PRId16, *(int16_t *)d); break;
      case mxINT32_CLASS:
        snprintf(skey, sizeof(skey), "%" PRId32, *(int32_t *)d); break;
      case mxINT64_CLASS:
        snprintf(skey, sizeof(skey), "%" PRId64, *(int64_t *)d); break;

      case mxUINT8_CLASS:
        snprintf(skey, sizeof(skey), "%" PRIu8, *(uint8_t *)d); break;
      case mxUINT16_CLASS:
        snprintf(skey, sizeof(skey), "%" PRIu16, *(uint16_t *)d); break;
      case mxUINT32_CLASS:
        snprintf(skey, sizeof(skey), "%" PRIu32, *(uint32_t *)d); break;
      case mxUINT64_CLASS:
        snprintf(skey, sizeof(skey), "%" PRIu64, *(uint64_t *)d); break;

      case mxSINGLE_CLASS:
        snprintf(skey, sizeof(skey),
                 (strcmp(kind, "float") && strcmp(kind, "double")) ?
                 "%.0f" : "%g", *(float *)d); break;
      case mxDOUBLE_CLASS:
        snprintf(skey, sizeof(skey),
                 (strcmp(kind, "float") && strcmp(kind, "double")) ?
                 "%.0f" : "%g", *(double *)d); break;

      default:
        snprintf(emsg, emsgsize, "bad numeric value for %s %s", kind, key);
        return -1;
    }

    s = json_bufcat(json, jend, jlen, 0, skey);
    return s;
  }

fatal:
  snprintf(emsg, emsgsize, "internal error at '%s'", key);
  return -1;
}


/* --- s2j_value ----------------------------------------------------------- */

int
s2j_value(const mxArray *value, char **json, char **jend, size_t *jlen)
{
  char skey[32];
  const char *field;
  mxArray *svalue;
  char *t;
  int i, n, s;

  if (mxIsStruct(value)) {
    if ((s = json_bufcat(json, jend, jlen, 0, "{"))) return s;
    n = mxGetNumberOfFields(value);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(value, i);
      svalue = mxGetFieldByNumber(value, 0, i);

      if (i && (s = json_bufcat(json, jend, jlen, 0, ","))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, field))) return s;
      if ((s = json_bufcat(json, jend, jlen, 0, "\":"))) return s;

      s = s2j_value(svalue, json, jend, jlen);
      if (s) return s;
    }
    if ((s = json_bufcat(json, jend, jlen, 0, "}"))) return s;
    return 0;

  } else if (mxIsCell(value)) {
    n = mxGetNumberOfElements(value);

    if ((s = json_bufcat(json, jend, jlen, 0, "["))) return s;
    for(i = 0; i < n; i++) {
      svalue = mxGetCell(value, i);
      if (!svalue) break;

      if (i && (s = json_bufcat(json, jend, jlen, 0, ","))) return s;

      s = s2j_value(svalue, json, jend, jlen);
      if (s) return s;
    }
    if ((s = json_bufcat(json, jend, jlen, 0, "]"))) return s;

    return 0;

  } else if (mxIsEmpty(value)) {
    return json_bufcat(json, jend, jlen, 0, tok_null);

  } else if (mxIsChar(value)) {
    t = mxArrayToString(value);

    if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;
    if (t && (s = json_bufcat(json, jend, jlen, 1, t))) return s;
    if ((s = json_bufcat(json, jend, jlen, 0, "\""))) return s;

    mxFree(t);
    return 0;

  } else if (mxIsNumeric(value)) {
    snprintf(skey, sizeof(skey), "%g", mxGetScalar(value));

    return json_bufcat(json, jend, jlen, 0, skey);
  }

  return -1;
}


/* --- json_bufcat --------------------------------------------------------- */

static int
json_bufcat(char **buf, char **end, size_t *len, int quoted, const char *data)
{
  size_t avail;
  const char *s;
  char *d;

  s = data;
  while(*s) {
    avail = *len - (*end - *buf);
    if (quoted) {
      for(d = *end; *s && avail; avail--) {
        if (*s == '"' || *s == '\\') {
          if (avail < 2) {
            avail = 0;
            break;
          }
          *d++ = '\\';
          avail--;
        }
        *d++ = *s++;
      }
    } else {
      for(d = *end; *s && avail; avail--)
        *d++ = *s++;
    }

    if (!avail) {
      char *newbuf = (char *)realloc(*buf, *len << 1);
      if (!*newbuf) return ENOMEM;

      d = newbuf + (d - *buf);
      *buf = newbuf;
      *len <<= 1;
    }
    *d = 0;
    *end = d;
  }

  return 0;
}
